
package java_cup.interpreter;

import java.util.LinkedList;
import java.util.ListIterator;
/** Class containing the implementation of the dynamic semantics of EGO 
 *@author Andi Bejleri
 *@version 0.1
 */
class Exec{
    
    /** creates a new object from an existing one, location. The list of
     *	methods and the address of the super object are copied from the
     *	cloned object to the newly created location.
     *@param location is the cloned object.
     *@return the new object created.
    */ 
    static Value exec_clone(Value location) throws BadFormedExpressionException{ 
	
	Object_descriptor odescr = null;
	Value ret = null;
	
	if( location.object != null )
	    if( location.object.compareTo("Object") == 0 ){
		/* creating an empty object */
		odescr = new Object_descriptor(null, new LinkedList());
		ret = new Value(odescr);
		/* debugging prints */
		System.out.println("New Object: " + ret.toString()+'\n');
		return ret;
	    }
	    else
		return new Value(new NotEvaluatedExpr("clone", location));
	else if( location.expr != null )
	    /* the value returned is part of a method body so it is not evaluated */
	    return new Value(new NotEvaluatedExpr("clone", location)); 
	
	else if( location.method != null )
	    /* The value is a method */
	    throw new BadFormedExpressionException("location.method != null ");
	
	else{
	    odescr = location.object_descriptor;
	    Object_descriptor obj = new Object_descriptor(odescr.deleg, odescr.copy_methods());
	    ret = new Value(obj);
	    /* debugging prints */
	    System.out.println("New Object: " + ret.toString() + '\n');
	    
	    return ret;
	}
    }
    
    /** adds a new method to the receiver object. It also changes the body of method of the receiver if the method name is already defined.
     *@param location the location of the receiver that will be added the method. 
     *@param mdescr the method added.
     *@return the result returned is the modified location of the receiver.
    */
    static Value exec_addMethod(Value location, Method_descriptor mdescr) throws BadFormedExpressionException{
	
	if( location.object != null || location.expr != null  )
	    /* the value returned is part of a method body so it is not evaluated */
	    return new Value(new NotEvaluatedExpr("addMethod", location, mdescr));
	
	else if( location.object_descriptor != null ){
	    /* debbuging prints */
	    System.out.println("The method to add: " + mdescr.toString() + '\n');
	
	    if(!location.object_descriptor.methods.isEmpty()){
	   
		/* the list of methods is not empty so we have to distinguish if it is a method body change or adding a new method */
		ListIterator iterator = location.object_descriptor.methods.listIterator(0);
		Method_descriptor meth = null;
		do{
		    meth = (Method_descriptor)iterator.next();
		    if( meth.meth_name.compareTo(mdescr.meth_name) == 0 ){
			/* Change the body of a method  */
			iterator.remove();
			location.object_descriptor.methods.add(mdescr);
			/*debbuging prints*/
			System.out.println("New object: "+location.toString() + '\n');

			return location;
		    }
		}while(iterator.hasNext());
	    }
	    /* Adding a new method  */
	    location.object_descriptor.methods.add(mdescr);
	    
	    /* debugging prints */
	    System.out.println("New Object: " +location.toString() + '\n');
	    
	    return location;
	}
	else
	    throw new BadFormedExpressionException();
    }
    /** changes the reference to the super object of the
     *	receiver object. 
     *@param location the location that will delegate to @see toDelegate
     *@param toDelegate the location to be delegated
     *@return the result of the reduction is the modified location of the receiver. 
    */
    static Value exec_delegate(Value location, Value toDelegate) throws BadFormedExpressionException{
	
	if( location.object != null || toDelegate.object != null || location.expr != null || toDelegate.expr != null )
	    /* the value returned is part of a method body so it is not evaluated */
	    return new Value(new NotEvaluatedExpr("delegate", location, toDelegate));
	else if( location.object_descriptor != null && toDelegate.object_descriptor != null ){
	    location.object_descriptor.deleg = toDelegate;
	    /* debugging prints */
	    System.out.println("New Object: "+location.toString() + '\n');
	    
	    return location;
	}
	else
	    throw new BadFormedExpressionException();
    }
    
    /** invokes a method on an object. The method is
     *owned by the receiver and is linear or nonlinear. As the type system does not
     *allow another call to a linear method we remove it from the
     *store. The location @see location is passed as an argument to the
     *method because self is not a free variable in the lambda
     *expression. 
     *@param location the receiver object
     *@param method_name method name to be invoked
     *@return the result of the reduction is a method
     *apply with @see location as an argument and a store without the
     *method @see method_name in it if @see method_name is linear or an unchanged store it @see method_name is nonlinear
     */
    static Value exec_invk(Value location, String method_name) throws BadFormedExpressionException{
	
	if( location.expr != null || location.object != null )
	    /* the value returned is part of a method body so it is not evaluated */
	    return new Value(new NotEvaluatedExpr("invk", location, method_name));
	
	else if( location.object_descriptor != null ){
	    
	    Value ret = null;
	    ListIterator iterator =location.object_descriptor.methods.listIterator(0);
	    Method_descriptor method = null;
	    Value theMethod = null;
	    do{
		/* scan the methods list */
		method = (Method_descriptor)iterator.next();
		
		if( method.meth_name.compareTo(method_name) == 0 ){
		    
		    theMethod = (Value)(method.method_body).clone();
		    if( method.method_body.method.linearity == 'l')
			/* The method is removed as it is linear */
			iterator.remove();
		    break;
		}
	    }while(iterator.hasNext());
	    
	    ret = exec_apply(theMethod, location);
	    /* debugging prints */
	    System.out.println("Method after invoked: "+ret.toString() + '\n');
	    
	    return ret;
	}
	else
	    throw new BadFormedExpressionException();
    }
    
    /** shows how a method is applied to its arguments.
     * @param method the method to be applied to its argument @see argument
     * @param argument the argument to be applied to @see method
     * @return the result of replacing the argument in the method @see method by @see arguments in
     * the method body of @see method  
    */
    static Value exec_apply( Value method, Value argument ) throws BadFormedExpressionException{
	
	if (method.expr != null || argument.expr != null || method.object != null || argument.object != null )
		return new Value(new NotEvaluatedExpr("apply", method, argument));
	else if( method.method != null && argument.object_descriptor != null ){
	    
	    substitute(method, method.method.arg, argument);
	    System.out.println("Method after substitution: " + method.toString() + '\n');
	    /* execute the unevaluated expression */
	    Value val = exec(method.method.code);
	    System.out.println("Value: " + val.toString() + '\n');
	    return val;
	}
	else
	    throw new BadFormedExpressionException();
    }

    /** does not effect the memory. It changes the linearity
     *of an object from linear to non-linear.
     * @param location location that will change the linearity
     * @return the result of the reduction is the location passed as argument for the expression.
    */
    static Value exec_changeLinearity( Value location ) throws BadFormedExpressionException{

	if( location.expr != null || location.object != null )
	    /* the value returned is part of a method body so it is not evaluated */
	    return new Value(new NotEvaluatedExpr("change", location));
	/* debugging prints */
	else if( location.object_descriptor != null ){
	    System.out.println("Object: "+location.toString() + '\n');
	    return location;
	}
	else
	    throw new BadFormedExpressionException();
    }

    /** substitutes the variables with a value in an unevaluated expression.
     * @param method_body the expression where it will take place the substitution.
     * @param arg the variable to be substituted
     * @param argument the location that will substitute @see arg
     */
    static void substitute( Value method_body, String arg, Value argument ){
	
	if( method_body.object_descriptor != null)
	    ;
	else if( method_body.object != null )
	    if( method_body.object.compareTo(arg) == 0 )
		
		method_body = argument;
	    else
		;
	else 
	    if( method_body.method != null )
	    
		if( method_body.method.code.object != null )
		    
		    if( method_body.method.code.object.compareTo(arg) == 0 )
			/* match variable */
			method_body.method.code = argument;
		    else
			/* no match */
			;	    
		else 
		    if(method_body.method.code.expr != null )
			if( method_body.method.code.expr.operation.compareTo("clone") == 0 )
			    /* Case clone */
			    substitute(method_body.method.code, arg, argument);
			else if( method_body.method.code.expr.operation.compareTo("delegate") == 0 ){
			    /* Case delegate */
			    substitute(method_body.method.code, arg, argument);
			    substitute((Value)method_body.method.code.expr.op_par, arg, argument);
			}
			else if( method_body.method.code.expr.operation.compareTo("addMethod") == 0 ){
			    /* Case addMethod */
			    substitute(method_body.method.code, arg, argument);
			    substitute( ((Method_descriptor)method_body.method.code.expr.op_par).method_body, arg, argument);
			}
			else if( method_body.method.code.expr.operation.compareTo("apply") == 0 ){
			    /* Case apply  */
			    substitute(method_body.method.code, arg, argument);
			    substitute((Value)method_body.method.code.expr.op_par, arg, argument);
			}
			else if( method_body.method.code.expr.operation.compareTo("invk") == 0 )
			    /* Case invk */
			    substitute(method_body.method.code, arg, argument);
			else //if( method_body.expr.operation.compareTo("change") == 0 )
			    /* Case change_linearity */
			    substitute(method_body.method.code, arg, argument);
		    else if( method_body.method.code.method != null )
			substitute( method_body.method.code, arg, argument );
		    else ;	
	    else    
		if(method_body.expr.expr.object != null )
		   
		    if(method_body.expr.expr.object.compareTo(arg) == 0 )
			method_body.expr.expr = argument; 
		    else
			;
		
		else
		    if( method_body.expr.operation.compareTo("clone") == 0 )
			/* Case clone */
			substitute( method_body.expr.expr, arg, argument );
		    else if( method_body.expr.operation.compareTo("delegate") == 0 ){
			/* Case delegate */
			substitute( method_body.expr.expr, arg, argument );
			substitute((Value)method_body.expr.op_par, arg, argument);
		    }
		    else if( method_body.expr.operation.compareTo("addMethod") == 0 ){
			/* Case addMethod */
			substitute( method_body.expr.expr, arg, argument );
			substitute( ((Method_descriptor)method_body.expr.op_par).method_body, arg, argument);
		    }
		    else if( method_body.expr.operation.compareTo("apply") == 0 ){
			/* Case apply  */
			substitute( method_body.expr.expr, arg, argument );
			substitute((Value)method_body.expr.op_par, arg, argument);
		    }
		    else if( method_body.expr.operation.compareTo("invk") == 0 )
			/* Case invk */
			substitute( method_body.expr.expr, arg, argument );
		    else //if( method_body.expr.operation.compareTo("change") == 0 )
			/* Case change_linearity */
			substitute( method_body.expr.expr, arg, argument );		
    }
    
    /** evaluates a not evaluated expression.
     *@param method_body the not evaluated expression
     *@return the evaluated expression
     */
    static Value exec( Value method_body) throws BadFormedExpressionException{
	
	if( method_body.object_descriptor != null )
	    /* body is a single object */
	    return method_body;
	else if( method_body.method != null )
	    /* body is a method */
	    return method_body;
	else
	    if( method_body.expr != null )
		/* body is an expression that have to reduced in a value */
		if( method_body.expr.operation.compareTo("clone") == 0 )
		    /* clone */
		    return exec_clone(exec(method_body.expr.expr));
		else if( method_body.expr.operation.compareTo("delegate") == 0 )
		    /* delegate */
		    return exec_delegate( exec(method_body.expr.expr), exec((Value)method_body.expr.op_par));
		else if( method_body.expr.operation.compareTo("addMethod") == 0 )
		    /* addMethod */
		    return exec_addMethod( exec(method_body.expr.expr), new Method_descriptor( ((Method_descriptor)method_body.expr.op_par).meth_name, exec( ((Method_descriptor)method_body.expr.op_par).method_body ) ) );
		else if( method_body.expr.operation.compareTo("apply") == 0 )
		    /* apply */
		    return exec_apply( exec(method_body.expr.expr), exec( (Value)method_body.expr.op_par) );
		else if( method_body.expr.operation.compareTo("invk") == 0 )
		    /* invk */
		    return exec_invk( exec(method_body.expr.expr), (String)method_body.expr.op_par );
		else //if( method_body.expr.operation.compareTo("change") == 0 )
		    /* change_linearity */
		    return exec_changeLinearity( exec(method_body.expr.expr) );
	    else 
		/* variable */
		return method_body;
    }
}
